#include <algorithm>
#include <cmath>
#include <msp/core/maputils.h>
#include <msp/core/raii.h>
#include <msp/strings/format.h>
#include <msp/time/units.h>
#include <msp/time/utils.h>
#include "beamgate.h"
#include "block.h"
#include "catalogue.h"
#include "driver.h"
#include "layout.h"
#include "simplecontroller.h"
#include "speedquantizer.h"
#include "timetable.h"
#include "trackcircuit.h"
#include "trackiter.h"
#include "tracktype.h"
#include "train.h"
#include "trainrouter.h"
#include "vehicle.h"
#include "vehicletype.h"

using namespace std;
using namespace Msp;

namespace R2C2 {

Train::Train(Layout &l, const VehicleType &t, unsigned a, const string &p):
	layout(l),
	loco_type(t),
	address(a),
	protocol(p),
	preceding_train(0),
	allocator(*this),
	advancing(false),
	controller(new SimpleController),
	current_speed_step(0),
	speed_changing(false),
	reverse(false),
	functions(0),
	pure_speed(false),
	speed_quantizer(0),
	accurate_position(false),
	overshoot_dist(false)
{
	if(!loco_type.is_locomotive())
		throw invalid_argument("Train::Train");

	unsigned speed_steps = layout.get_driver().get_protocol_speed_steps(protocol);
	if(speed_steps)
		speed_quantizer = new SpeedQuantizer(speed_steps);

	vehicles.push_back(new Vehicle(layout, loco_type));
	vehicles.back()->set_train(this);

	layout.add_train(*this);

	loco_id = layout.get_driver().add_loco(address, protocol, loco_type);
	layout.get_driver().signal_loco_speed.connect(sigc::mem_fun(this, &Train::loco_speed_event));
	layout.get_driver().signal_loco_function.connect(sigc::mem_fun(this, &Train::loco_func_event));

	layout.get_driver().signal_halt.connect(sigc::mem_fun(this, &Train::halt_event));

	controller->signal_control_changed.connect(sigc::mem_fun(this, &Train::control_changed));

	allocator.signal_advanced.connect(sigc::mem_fun(this, &Train::advanced));
	allocator.signal_rear_advanced.connect(signal_rear_advanced);
}

Train::~Train()
{
	delete controller;
	for(vector<Vehicle *>::iterator i=vehicles.begin(); i!=vehicles.end(); ++i)
		delete *i;
	layout.remove_train(*this);
}

void Train::set_name(const string &n)
{
	name = n;

	signal_name_changed.emit(name);
}

void Train::add_vehicle(const VehicleType &vt)
{
	Vehicle *veh = new Vehicle(layout, vt);
	vehicles.back()->attach_back(*veh);
	vehicles.push_back(veh);
	veh->set_train(this);
	signal_vehicle_added.emit(vehicles.size()-1, *veh);
}

void Train::remove_vehicle(unsigned i)
{
	if(i>=vehicles.size())
		throw out_of_range("Train::remove_vehicle");
	if(i==0)
		throw logic_error("can't remove locomotive");

	Vehicle *veh = vehicles[i];
	vehicles.erase(vehicles.begin()+i);
	veh->detach_front();
	if(i<vehicles.size())
	{
		veh->detach_back();
		vehicles[i-1]->attach_back(*vehicles[i]);
	}
	signal_vehicle_removed.emit(i, *veh);
	delete veh;
}

unsigned Train::get_n_vehicles() const
{
	return vehicles.size();
}

Vehicle &Train::get_vehicle(unsigned i)
{
	if(i>=vehicles.size())
		throw out_of_range("Train::get_vehicle");
	return *vehicles[i];
}

const Vehicle &Train::get_vehicle(unsigned i) const
{
	if(i>=vehicles.size())
		throw out_of_range("Train::get_vehicle");
	return *vehicles[i];
}

void Train::set_control(const string &n, float v)
{
	controller->set_control(n, v);
}

void Train::set_function(unsigned func, bool state)
{
	if(!loco_type.get_functions().count(func))
		throw invalid_argument("Train::set_function");
	layout.get_driver().set_loco_function(loco_id, func, state);
}

float Train::get_control(const string &ctrl) const
{
	return controller->get_control(ctrl).value;
}

float Train::get_speed() const
{
	return controller->get_speed();
}

float Train::get_quantized_speed() const
{
	if(speed_quantizer)
		return speed_quantizer->quantize_speed(controller->get_speed());
	else
		return controller->get_speed();
}

float Train::get_maximum_speed() const
{
	float ms = 0;
	for(vector<Vehicle *>::const_iterator i=vehicles.begin(); i!=vehicles.end(); ++i)
	{
		float vms = (*i)->get_type().get_maximum_speed();
		if(ms<=0 || (vms>0 && vms<ms))
			ms = vms;
	}
	return ms;
}

bool Train::get_function(unsigned func) const
{
	return (functions>>func)&1;
}

void Train::add_ai(TrainAI &ai)
{
	ais.push_back(&ai);
	ai.signal_event.connect(sigc::bind<0>(signal_ai_event, sigc::ref(ai)));
}

void Train::remove_ai(TrainAI &ai)
{
	list<TrainAI *>::iterator i = find(ais.begin(), ais.end(), &ai);
	if(i!=ais.end())
		ais.erase(i);
}

void Train::ai_message(const TrainAI::Message &msg)
{
	for(list<TrainAI *>::iterator i=ais.begin(); i!=ais.end(); ++i)
		(*i)->message(msg);
}

bool Train::place(const BlockIter &block)
{
	if(!block)
		throw invalid_argument("Train::place");
	if(controller->get_speed())
		throw logic_error("moving");

	accurate_position = false;
	last_entry_block = BlockIter();

	if(allocator.start_from(block))
	{
		if(reverse)
			vehicles.front()->place(block.reverse().track_iter(), VehiclePlacement::FRONT_BUFFER);
		else
			vehicles.back()->place(block.track_iter(), VehiclePlacement::BACK_BUFFER);
		return true;
	}
	else
	{
		unplace();
		return false;
	}
}

void Train::unplace()
{
	if(controller->get_speed())
		throw logic_error("moving");

	allocator.clear();
	accurate_position = false;
	last_entry_block = BlockIter();

	for(vector<Vehicle *>::iterator i=vehicles.begin(); i!=vehicles.end(); ++i)
		(*i)->unplace();
}

void Train::stop_at(Block *block)
{
	allocator.stop_at(block);
}

bool Train::is_block_critical(const Block &block) const
{
	if(allocator.is_block_current(block))
		return true;

	BlockIter i = check_critical_blocks(&block);
	return i.block()==&block;
}

BlockIter Train::get_last_critical_block() const
{
	return check_critical_blocks(0);
}

BlockIter Train::check_critical_blocks(const Block *check) const
{
	if(allocator.empty())
		return BlockIter();

	BlockIter i = allocator.last_current();

	if(controller->get_speed()==0)
		return i;

	BlockIter last = allocator.last();
	if(i.block()==last.block())
		return i;

	float margin = 10*layout.get_catalogue().get_scale();
	float min_dist = controller->get_braking_distance()*1.3+margin;

	float dist = 0;
	bool sensor_seen = false;
	i = i.next();
	while(i.block()!=last.block() && i.block()!=check)
	{
		dist += i->get_path_length(i.entry());

		if(i->get_sensor_address())
			sensor_seen = true;

		if(dist>min_dist && sensor_seen)
			break;

		BlockIter j = i.next();
		if(j->get_train()!=this)
			break;

		i = j;
	}

	return i;
}

void Train::refresh_blocks_from(Block &block)
{
	if(is_block_critical(block))
		allocator.rewind_to(*get_last_critical_block().next());
	else
		allocator.rewind_to(block);
}

float Train::get_reserved_distance() const
{
	if(allocator.empty())
		return 0;

	float margin = 0;
	TrackIter next = allocator.last().next().track_iter();
	if(next && next->get_type().is_turnout())
		margin = 15*layout.get_catalogue().get_scale();

	return max(get_reserved_distance_until(0)-margin, 0.0f);
}

void Train::tick(const Time::TimeDelta &dt)
{
	if(stop_timeout)
	{
		stop_timeout = max(stop_timeout-dt, Time::zero);
		if(stop_timeout<=Time::zero)
			allocator.set_active(false);
	}

	travel_time += dt;

	Driver &driver = layout.get_driver();

	allocator.tick();

	bool intent_to_move = false;
	for(list<TrainAI *>::iterator i=ais.begin(); i!=ais.end(); ++i)
	{
		(*i)->tick(dt);
		if((*i)->has_intent_to_move())
			intent_to_move = true;
	}

	controller->tick(dt);
	float speed = controller->get_speed();
	bool moving = speed>0;

	if(controller->get_reverse()!=reverse)
	{
		reverse = controller->get_reverse();
		bool r = reverse;
		if(loco_type.get_swap_direction())
			r = !r;
		driver.set_loco_reverse(loco_id, r);

		allocator.reverse();
		last_entry_block = BlockIter();
	}

	if(speed_quantizer)
	{
		unsigned speed_step = speed_quantizer->find_speed_step(speed);
		if(speed_step!=current_speed_step && !speed_changing && !driver.is_halted() && driver.get_power())
		{
			speed_changing = true;
			driver.set_loco_speed(loco_id, speed_step);

			pure_speed = false;
		}

		speed = speed_quantizer->get_speed(current_speed_step);
	}

	if(moving)
	{
		if(!allocator.is_active())
			allocator.set_active(true);

		Vehicle &vehicle = *(reverse ? vehicles.back() : vehicles.front());

		float d = speed*(dt/Time::sec);
		Block &block = vehicle.get_placement().get_position(reverse ? VehiclePlacement::BACK_AXLE : VehiclePlacement::FRONT_AXLE)->get_block();
		if(allocator.is_block_current(block))
		{
			SetFlag setf(advancing);
			vehicle.advance(reverse ? -d : d);
		}
		else if(accurate_position)
		{
			overshoot_dist += d;
			if(overshoot_dist>40*layout.get_catalogue().get_scale())
			{
				layout.emergency(&block, name+" has not arrived at sensor");
				accurate_position = false;
			}
		}
	}
	else if(intent_to_move && !allocator.is_active())
		allocator.set_active(true);
	else if(allocator.is_active() && !intent_to_move && !stop_timeout)
		stop_timeout = 2*Time::sec;
}

void Train::save(list<DataFile::Statement> &st) const
{
	st.push_back((DataFile::Statement("name"), name));

	const Catalogue &cat = layout.get_catalogue();
	for(vector<Vehicle *>::const_iterator i=vehicles.begin(); i!=vehicles.end(); ++i)
		if(i!=vehicles.begin())
			st.push_back((DataFile::Statement("vehicle"), cat.get_name(&(*i)->get_type())));

	if(speed_quantizer)
	{
		DataFile::Statement ss("quantized_speed");
		speed_quantizer->save(ss.sub);
		st.push_back(ss);
	}

	{
		DataFile::Statement ss("blocks");
		allocator.save(ss.sub);
		st.push_back(ss);
	}

	const VehicleType::FunctionMap &func_map = loco_type.get_functions();
	for(VehicleType::FunctionMap::const_iterator i=func_map.begin(); i!=func_map.end(); ++i)
		st.push_back((DataFile::Statement("function"), i->first, static_cast<bool>(functions&(1<<i->first))));

	// XXX Need more generic way of saving AI state
	for(list<TrainAI *>::const_iterator i=ais.begin(); i!=ais.end(); ++i)
	{
		if(TrainRouter *router = dynamic_cast<TrainRouter *>(*i))
		{
			DataFile::Statement ss("router");
			router->save(ss.sub);
			st.push_back(ss);
		}
		else if(Timetable *timetable = dynamic_cast<Timetable *>(*i))
		{
			DataFile::Statement ss("timetable");
			timetable->save(ss.sub);
			st.push_back(ss);
		}
	}
}

void Train::control_changed(const Controller::Control &ctrl)
{
	signal_control_changed.emit(ctrl.name, ctrl.value);
}

void Train::loco_speed_event(unsigned id, unsigned speed, bool rev)
{
	if(id==loco_id)
	{
		current_speed_step = speed;
		bool r = reverse;
		if(loco_type.get_swap_direction())
			r = !r;
		if(rev!=r)
			layout.get_driver().set_loco_reverse(loco_id, r);
		speed_changing = false;
		pure_speed = false;
	}
}

void Train::loco_func_event(unsigned id, unsigned func, bool state)
{
	if(id==loco_id)
	{
		if(state)
			functions |= 1<<func;
		else
			functions &= ~(1<<func);

		signal_function_changed.emit(func, state);
	}
}

void Train::advanced(Block &block, Sensor *sensor)
{
	if(!sensor || sensor==block.get_sensor())
		signal_advanced.emit(block);

	if(!sensor)
		return;

	if(sensor==block.get_sensor())
	{
		if(last_entry_block && pure_speed && speed_quantizer)
		{
			float travel_distance = 0;

			for(BlockIter i=last_entry_block; &*i!=&block; i=i.next())
				travel_distance += i->get_path_length(i.entry());

			if(travel_distance>0)
			{
				float travel_time_secs = travel_time/Time::sec;

				if(travel_time_secs>=2)
					speed_quantizer->learn(current_speed_step, travel_distance/travel_time_secs, travel_time_secs);
			}
		}

		last_entry_block = allocator.iter_for(block);
		travel_time = Time::zero;
		if(!layout.get_driver().is_halted())
		{
			pure_speed = true;
			accurate_position = true;
		}
		overshoot_dist = 0;

		if(!advancing && vehicles.front()->is_placed())
		{
			TrackIter track = last_entry_block.track_iter();
			if(reverse)
			{
				track = track.flip();
				vehicles.back()->place(track, VehiclePlacement::BACK_AXLE);
			}
			else
				vehicles.front()->place(track, VehiclePlacement::FRONT_AXLE);
		}
	}
	else if(BeamGate *gate = dynamic_cast<BeamGate *>(sensor))
	{
		if(!advancing && vehicles.front()->is_placed())
		{
			TrackIter track = allocator.iter_for(block).track_iter();
			for(; (track && &track->get_block()==&block); track=track.next())
				if(track.track()==gate->get_track())
				{
					if(reverse)
						track = track.reverse();
					float offset = gate->get_offset_from_endpoint(track.entry());
					if(reverse)
						vehicles.back()->place(TrackOffsetIter(track, offset), VehiclePlacement::BACK_BUFFER);
					else
						vehicles.front()->place(TrackOffsetIter(track, offset), VehiclePlacement::FRONT_BUFFER);
					break;
				}
		}
	}
}

void Train::halt_event(bool h)
{
	if(h)
		accurate_position = false;
}

float Train::get_reserved_distance_until(const Block *until_block) const
{
	if(allocator.empty())
		return 0;

	Vehicle &veh = *(reverse ? vehicles.back() : vehicles.front());

	TrackOffsetIter track = veh.get_placement().get_position(reverse ? VehiclePlacement::BACK_AXLE : VehiclePlacement::FRONT_AXLE);
	if(!track)  // XXX Probably unnecessary
		return 0;

	if(&track->get_block()==until_block)
		return 0;

	// Account for the vehicle's offset on its current track
	float result = track.offset();
	if(reverse)
		track = track.reverse();
	else
		result = track->get_path_length()-result;
	result -= veh.get_type().get_length()/2;

	BlockIter block = track.block_iter();

	// Count remaining distance in the vehicle's current block
	for(track=track.next(); &track->get_block()==&*block; track=track.next())
		result += track->get_path_length();

	const BlockIter &last = allocator.last();
	if(&*block==&*last)
		return result;

	// Count any remaining blocks
	for(block=block.next(); (&*block!=until_block && block->get_train()==this); block=block.next())
	{
		result += block->get_path_length(block.entry());

		if(&*block==&*last)
			break;
	}

	return result;
}


Train::Loader::Loader(Train &t):
	DataFile::ObjectLoader<Train>(t),
	prev_block(0),
	blocks_valid(true)
{
	add("blocks",      &Loader::blocks);
	add("function",    &Loader::function);
	add("name",        &Loader::name);
	add("quantized_speed",  &Loader::quantized_speed);
	add("router",      &Loader::router);
	add("timetable",   &Loader::timetable);
	add("vehicle",     &Loader::vehicle);
}

void Train::Loader::finish()
{
	if(!obj.allocator.empty())
	{
		TrackIter track = obj.allocator.first().track_iter();
		float offset = 2*obj.layout.get_catalogue().get_scale();
		obj.vehicles.back()->place(TrackOffsetIter(track, offset), VehiclePlacement::BACK_BUFFER);
	}
}

void Train::Loader::blocks()
{
	load_sub(obj.allocator);
}

void Train::Loader::function(unsigned f, bool s)
{
	obj.set_function(f, s);
}

void Train::Loader::name(const string &n)
{
	obj.set_name(n);
}

void Train::Loader::quantized_speed()
{
	if(obj.speed_quantizer)
		load_sub(*obj.speed_quantizer);
}

void Train::Loader::router()
{
	TrainRouter *rtr = new TrainRouter(obj);
	load_sub(*rtr);
}

void Train::Loader::timetable()
{
	Timetable *ttbl = new Timetable(obj);
	load_sub(*ttbl, obj.layout);
}

void Train::Loader::vehicle(const string &n)
{
	const VehicleType &vtype = obj.layout.get_catalogue().get<VehicleType>(n);
	Vehicle *veh = new Vehicle(obj.layout, vtype);
	obj.vehicles.back()->attach_back(*veh);
	obj.vehicles.push_back(veh);
	veh->set_train(&obj);
}

} // namespace R2C2
